# coding: utf-8
# Copyright (c) Qotto, 2018

import json
import os
import yaml

import logging
from ..event.base import BaseEvent
from ..event.generic import GenericEvent
from .base import BaseEventSerializer
from avro.datafile import DataFileWriter, DataFileReader
from avro.io import DatumWriter, DatumReader
from avro.schema import NamedSchema, Parse
from io import BytesIO

from typing import Dict, Type

__all__ = [
    'AvroEventSerializer',
]


class AvroEventSerializer(BaseEventSerializer):
    AVRO_SCHEMA_FILE_EXTENSION = 'avsc.yaml'

    def __init__(self, settings: object) -> None:
        self.logger = logging.getLogger(__name__)
        self._schemas: Dict[str, NamedSchema] = dict()
        self._events: Dict[str, Type[BaseEvent]] = dict()

        if not hasattr(settings, 'AVRO_SCHEMAS_FOLDER'):
            raise Exception('Missing AVRO_SCHEMAS_FOLDER config')
        self.scan_folder(schemas_folder=settings.AVRO_SCHEMAS_FOLDER)

    def scan_folder(self, schemas_folder: str) -> None:
        """
        Searches Avro schema files in the specified folder.
        The schema files should have the avsc.yaml extension.
        """
        with os.scandir(schemas_folder) as entries:  # type: ignore
            for entry in entries:
                # ignores everything that is not a file
                if not entry.is_file():
                    continue
                # ignores hidden files
                if entry.name.startswith('.'):
                    continue
                # ignores everything files with wrong extension
                if not entry.name.endswith(f'.{self.AVRO_SCHEMA_FILE_EXTENSION}'):
                    continue
                self._extract_schema_from_file(entry.path)

    def _extract_schema_from_file(self, file_path: str) -> None:
        with open(file_path, 'r') as f:
            for s in yaml.load_all(f):
                avro_schema_data = json.dumps(s)
                avro_schema = Parse(avro_schema_data)
                if avro_schema.name in self._schemas:
                    raise Exception(
                        f"Avro schema {avro_schema.name} was defined more than once!")
                self._schemas[avro_schema.name] = avro_schema

    def register_event_class(self, event_class: Type[BaseEvent], event_name: str = None) -> None:
        """
        Registers an event class associated to a particular event name.
        """
        if event_name is None:
            event_name = event_class.__name__
        if event_name not in self._schemas:
            raise NameError(f"{event_name} event does not exist")
        self._events[event_name] = event_class

    def encode(self, event: BaseEvent) -> bytes:
        schema = self._schemas[event.name]
        output = BytesIO()
        writer = DataFileWriter(output, DatumWriter(), schema)
        writer.append(event.data)
        writer.flush()
        encoded_event = output.getvalue()
        writer.close()
        return encoded_event

    def decode(self, encoded_event: bytes) -> BaseEvent:
        reader = DataFileReader(BytesIO(encoded_event), DatumReader())
        schema = json.loads(reader.meta.get('avro.schema').decode('utf-8'))
        event_name = schema['name']
        event_data = next(reader)
        event_class = self._events.get(event_name, GenericEvent)

        return event_class.from_data(event_name, event_data)



from ..event.base import BaseEvent
from ..app.base import BaseApp

__all__ = [
    'BaseCommand',
]


class BaseCommand(BaseEvent):
    async def handle(self, app: BaseApp, corr_id: str):
        await self.execute(app=app, corr_id=corr_id)

    async def execute(self, app: BaseApp, corr_id: str):
        pass


class BaseCommandResult(BaseEvent):
    @property
    def error(self):
        return self.data['error']

    @property
    def message(self):
        return self.data['message']



from aiokafka import AIOKafkaConsumer
import logging
import sys
from typing import Any, Dict, Type, Callable, List
import asyncio
from ..serializer.base import BaseEventSerializer
from ..event.base import BaseEvent
from ..command.base import BaseCommand
from .base import BaseEventConsumer
from ..app.base import BaseApp

__all__ = [
    'KafkaConsumer'
]


class KafkaConsumer(BaseEventConsumer):

    def __init__(self, settings: object, app: BaseApp, serializer: BaseEventSerializer, event_topics: List[str], event_group: str, position: str) -> None:
        self.logger = logging.getLogger(__name__)

        if not hasattr(settings, 'KAFKA_BOOTSTRAP_SERVER'):
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')

        self.app = app
        self.event_topics = event_topics
        self.event_group = event_group
        self.position = position
        self.consumer = None
        self.serializer = serializer
        self.bootstrap_servers = settings.KAFKA_BOOTSTRAP_SERVER

    async def start(self):

        consumer_args: Dict[str, Any]
        consumer_args = {
            'loop': asyncio.get_event_loop(),
            'bootstrap_servers': [self.bootstrap_servers],
            'enable_auto_commit': False,
            'group_id': self.event_group,
            'value_deserializer': self.serializer.decode,
            'auto_offset_reset': self.position
        }

        try:
            self.consumer = AIOKafkaConsumer(
                *self.event_topics, **consumer_args)
        except Exception as e:
            self.logger.error(
                f"Unable to connect to the Kafka broker {self.bootstrap_servers} : {e}")
            raise e

        self.logger.info(
            f'Starting kafka consumer on topic {self.event_topics} with group {self.event_group}')
        try:
            await self.consumer.start()
        except Exception as e:
            self.logger.error(
                f'An error occurred while starting kafka consumer on topic {self.event_topics} with group {self.event_group}: {e}')
            sys.exit(1)

        try:
            async for msg in self.consumer:

                event = msg.value
                corr_id = event.correlation_id

                self.logger.info(
                    f"[CID:{corr_id}] Start handling {event.name}")
                await event.handle(app=self.app, corr_id=corr_id)
                self.logger.info(
                    f"[CID:{corr_id}] End handling {event.name}")

                if self.event_group is not None:
                    self.logger.debug(
                        f"[CID:{corr_id}] Commit Kafka transaction")
                    await self.consumer.commit()
        except Exception as e:
            self.logger.error(
                f'An error occurred while handling received message : {e}')
            raise e
        finally:
            await self.consumer.stop()



# coding: utf-8
# Copyright (c) Qotto, 2018

from base64 import b64encode
from datetime import datetime
from datetime import timezone
from secrets import token_urlsafe
from typing import Any
from pydoc import locate

__all__ = [
    'date2timestamp',
    'current_timestamp',
    'gen_correlation_id',
]


def date2timestamp(date: datetime) -> int:
    """
    Converts a UTC date to a UNIX timestamp, in milliseconds.

    >>> date2timestamp(datetime(1970, 1, 1, tzinfo=timezone.utc))
    0
    >>> date2timestamp(datetime(1970, 1, 2, tzinfo=timezone.utc))
    86400000
    """
    return round(date.timestamp()*1000)


def timestamp2date(timestamp: int) -> datetime:
    """
    Converts a UNIX timestamp in milliseconds to a UTC date.

    >>> timestamp2date(0)
    datetime.datetime(1970, 1, 1, 1, 0)
    >>> timestamp2date(24*3600*1000)
    datetime.datetime(1970, 1, 2, 1, 0)
    """
    return datetime.fromtimestamp(timestamp/1000.0)


def current_timestamp() -> int:
    """
    Returns current UNIX timestamp, in milliseconds.

    Example: checks that time flows in the right direction:

    >>> t1 = current_timestamp()
    >>> t2 = current_timestamp()
    >>> t1 <= t2
    True
    """
    return date2timestamp(datetime.now(timezone.utc))


def gen_correlation_id(prefix: str = None) -> str:
    """
    Generates a correlation ID that looks like `prefix:date:random`, where:

    - `prefix` is a fixed part that you can specify (any length)
    - `date` only depends on the current date (8 characters)
    - `random` is a random part  (4 characters)

    `date` and `random` are encoded as `[a-zA-Z0-9_-]`.
    """
    CORRELATION_ID_PREFIX_LENGTH = 6
    CORRELATION_ID_TOKEN_LENGTH = 3

    def ts2000res65536() -> bytes:
        """
        Converts current date to 6 bytes
        """
        ts_now = datetime.now(timezone.utc).timestamp()
        ts_2k = datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp()
        return int(65536 * (ts_now - ts_2k)).to_bytes(6, 'big')

    if prefix is None:
        prefix = token_urlsafe(CORRELATION_ID_PREFIX_LENGTH)
    date = b64encode(ts2000res65536(), b'_-').decode('ascii')
    random = token_urlsafe(CORRELATION_ID_TOKEN_LENGTH)
    return f'{prefix}:{date}:{random}'


def get_dynamic_instance(full_class_name: str):
    """
    Returns an instance of the specified class

    >>> instance = get_dynamic_instance('eventy.consumer.base.BaseEventConsumer')
    >>> type(instance)
    <class 'eventy.consumer.base.BaseEventConsumer'>
    """
    class_ = load_class(full_class_name)
    return class_()


def load_class(full_class_name: str):
    """
    Loads the specified class

    >>> load_class('eventy.consumer.base.BaseEventConsumer')
    <class 'eventy.consumer.base.BaseEventConsumer'>
    """
    return locate(full_class_name)



# coding: utf-8
# Copyright (c) Qotto, 2018

from ..event.base import BaseEvent

from typing import Any, Type

__all__ = [
    'BaseEventSerializer',
]


class BaseEventSerializer:

    def __init__(self, settings: object) -> None:
        pass

    def encode(self, event: BaseEvent) -> Any:
        raise NotImplementedError

    def decode(self, encoded_event: Any) -> BaseEvent:
        raise NotImplementedError

    def register_event_class(self, event_class: Type[BaseEvent], event_name: str = None) -> None:
        raise NotImplementedError



# coding: utf-8
# Copyright (c) Qotto, 2018

from ..utils import current_timestamp, gen_correlation_id
from typing import Any, Dict
from ..app.base import BaseApp

__all__ = [
    'BaseEvent'
]


class BaseEvent:
    def __init__(self, name: str, data: Dict[str, Any]) -> None:
        if data is None:
            data = dict()

        if 'schema_version' not in data:
            data['schema_version'] = '0.0.0'

        if 'timestamp' not in data:
            data['timestamp'] = current_timestamp()

        self.name = name
        self.data = data

    async def handle(self, app: BaseApp, corr_id: str):
        pass

    @property
    def context(self):
        return self.data['context']

    @property
    def schema_version(self):
        return self.data['schema_version']

    @property
    def correlation_id(self):
        return self.data['correlation_id']

    @property
    def timestamp(self):
        return self.data['timestamp']

    @classmethod
    def from_data(cls, event_name: str, event_data: Dict[str, Any]):
        raise NotImplementedError



# coding: utf-8
# Copyright (c) Qotto, 2018

import doctest
import os
import unittest

def load_tests(loader, tests, pattern):
    test_dir = os.path.dirname(__file__)
    top_dir = os.path.dirname(test_dir)
    package_tests = loader.discover(start_dir=test_dir, pattern='test_*.py', top_level_dir=top_dir)
    tests.addTests(package_tests)
    return tests

if __name__ == '__main__':
    unittest.main()
